with Win32.Winerror; use Win32.Winerror;
with Win32.Rpcdce; use Win32.Rpcdce;
with Win32.Winbase;
with System; use System;
with Ada.Unchecked_Deallocation;
with Ada.Unchecked_Conversion;

with Beep_Global; use Beep_Global;
with BeepCom; use BeepCom;

package body Factory is

   procedure Free is new Ada.Unchecked_Deallocation(Class_Factory, Class_Factory_Pointer);
   procedure Free_Vtbl is new Ada.Unchecked_Deallocation(Class_Factory_Vtbl, Class_Factory_Vtbl_Pointer);

   function To_Class_Factory_Pointer is
      new Ada.Unchecked_Conversion (System.Address, Class_Factory_Pointer);
   function To_PVOID is
      new Ada.Unchecked_Conversion (System.Address, Win32.PVOID);

   procedure Free ( Pointer : in System.Address ) is
      X : Class_Factory_Pointer;
   begin
      X := To_Class_Factory_Pointer(Pointer);
      Free(X);
   end Free;

   function QueryInterface  (
                This     : access Class_Factory;
                riid     : REFIID;
                ppvObject: access Win32.PVOID)
               return HRESULT is
   begin

      if riid.all = Beep_Global.IID_IUnknown then
         ppvObject.all := This.all'Address;
      elsif riid.all = Beep_Global.IID_IClassFactory then
         ppvObject.all := This.all'Address;
	else
         ppvObject.all := System.Null_Address;
         return E_NOINTERFACE;
	end if;

      This.m_cRef := This.m_cRef+1;

      return S_OK;
   end QueryInterface;

   function AddRef (
                This: access Class_Factory)
               return Win32.ULONG is
   begin
      This.m_cRef := This.m_cRef+1;

      return This.m_cRef;
   end AddRef;

   function Release (
                This: access Class_Factory)
               return Win32.ULONG is
   begin
	This.m_cRef := This.m_cRef-1;

	if This.m_cRef /=0 then
         return This.m_cRef;
      end if;

	Free(This.all'Address);

	return 0;
   end Release;

   function CreateInstance (
                This     : access Class_Factory;
                pUnkOuter: access IUnknown;
                riid     : REFIID;
                ppvObject: access Win32.PVOID)
               return HRESULT is
      pObject : Beep_CoClass_Pointer;
      hr       : unsigned_long := 16#80040111#;
      lResult  : Win32.LONG;
      refcount : Win32.ULONG;
   begin
      pObject := new Beep_CoClass;

      hr := pObject.lpvtbl.QueryInterface(pObject, riid, ppvObject);

      refcount := pObject.lpvtbl.Release(pObject);

      lResult := Win32.Winbase.InterlockedIncrement(g_cComponents'Access);

      return hr;
   end CreateInstance;

   function LockServer (
                This : access Class_Factory;
                fLock: Win32.BOOL)
               return HRESULT is
      use Win32;
      lResult : Win32.LONG;
   begin
	if fLock = Win32.TRUE then
         lResult := Win32.Winbase.InterlockedIncrement(g_cServer'Access);
	else
         lResult := Win32.Winbase.InterlockedDecrement(g_cServer'Access);
      end if;

      Can_Close;

      return S_OK;
   end LockServer;

end Factory;